/***************************************************************************
 * Copyright 2012 by
 *  + Christian-Albrechts-University of Kiel
 *    + Department of Computer Science
 *      + Software Engineering Group 
 *  and others.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 ***************************************************************************/
package kieker.webgui.common;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.List;

import kieker.analysis.AnalysisController;
import kieker.analysis.model.analysisMetaModel.MIDependency;
import kieker.analysis.model.analysisMetaModel.MIProject;
import kieker.analysis.model.analysisMetaModel.impl.MAnalysisMetaModelFactory;
import kieker.common.logging.Log;
import kieker.common.logging.LogFactory;

import org.primefaces.model.UploadedFile;

/**
 * This is a singleton class for the access to the file system. It makes sure
 * that the necessary directories for the execution of the application exist.
 * <b>Do not</b> remove directories created from this manager during runtime!
 * Directories are created during first access to the class.<br>
 * Currently
 * nearly all methods are synchronized. A fine grained synchronization should be
 * implemented in the future.
 * 
 * @author Nils Christian Ehmke
 */
public final class FileManager {

	private static final Log LOG = LogFactory.getLog(FileManager.class);
	private static final String ROOT_DIR = "data";
	private static final String PROJECT_DIR = FileManager.ROOT_DIR + File.separator + "projects";
	private static final String LIB_DIR = FileManager.ROOT_DIR + File.separator + "libraries";
	private static final String EXTENSION = ".kax";
	private static final String JAR_EXTENSION = ".jar";
	private static final int BUF_SIZE = 1024;
	private static final FileManager INSTANCE = new FileManager();
	private final MAnalysisMetaModelFactory factory;

	/**
	 * Creates a new instance of this class.
	 */
	private FileManager() {
		this.checkAndCreateDirectories();
		this.factory = new MAnalysisMetaModelFactory();
	}

	/**
	 * Checks whether all directories are available and creates them if necessary.
	 */
	private synchronized void checkAndCreateDirectories() {
		/*
		 * Make sure that the directories exist and create them if necessary.
		 */
		final File dirProj = new File(FileManager.PROJECT_DIR);
		final File dirLib = new File(FileManager.LIB_DIR);
		boolean couldCreated = true;

		if (!dirProj.exists()) {
			couldCreated &= dirProj.mkdirs();
		}
		if (!dirLib.exists()) {
			couldCreated &= dirLib.mkdirs();
		}

		if (!couldCreated) {
			FileManager.LOG.error("Could not create the necessary directories for the application");
		}
	}

	/**
	 * This method saves a given project using the name of the project as the
	 * project-directory. The project-directory must already exist.
	 * 
	 * @param project
	 *            The project to be stored.
	 * @return true iff the project-directory does already exist and the storage
	 *         was successful.
	 */
	public synchronized boolean saveProject(final MIProject project) {
		final String projectName = project.getName();

		final File dirProject = new File(FileManager.PROJECT_DIR + File.separator + projectName);

		/*
		 * Make sure that the directory for the project exists.
		 */
		if (!dirProject.exists()) {
			return false;
		} else {
			/*
			 * Try to save the project.
			 */
			final File fileProject = new File(dirProject, projectName + FileManager.EXTENSION);
			try {
				final AnalysisController controller = new AnalysisController(project);
				return controller.saveToFile(fileProject, projectName);
			} catch (final Exception ex) {
				FileManager.LOG.error("Could not save project '" + projectName + "'.");
				return false;
			}
		}
	}

	/**
	 * This method saves a given project using the name of the project as the
	 * project-directory. THe project-directory must not already exist.
	 * 
	 * @param project
	 *            The project to be stored.
	 * @return true iff the project-directory does not already exist and the
	 *         storage was successful.
	 */
	public synchronized boolean saveNewProject(final MIProject project) {
		final String projectName = project.getName();
		final File dirProject = new File(FileManager.PROJECT_DIR + File.separator + projectName);

		/*
		 * Make sure that the project does not already exist and create the
		 * project directory.
		 */
		if (dirProject.exists() || !dirProject.mkdir()) {
			return false;
		} else {
			/*
			 * The directory should exist now. Store the project.
			 */
			return this.saveProject(project);
		}
	}

	/**
	 * This method can be used to load all currently saved projects.
	 * 
	 * @return A list containing the loaded projects. If something went wrong,
	 *         an empty list will be returned, never null.
	 */
	public synchronized List<MIProject> loadAllProjects() {
		final List<MIProject> resultList = new ArrayList<MIProject>();

		/*
		 * Try to get all directories within the project directory.
		 */
		final File[] directories = new File(FileManager.PROJECT_DIR).listFiles();
		if (directories != null) {
			for (final File directory : directories) {
				if (directory.isDirectory()) {
					/*
					 * If there is a project file within the directory, we know
					 * the name of it.
					 */
					final File projectFile = new File(directory, directory.getName() + FileManager.EXTENSION);
					if (projectFile.exists()) {
						try {
							/*
							 * Try to load the project.
							 */
							final MIProject project = AnalysisController.loadFromFile(projectFile);
							if (project != null) {
								resultList.add(project);
							}
						} catch (final Exception ex) {
							FileManager.LOG.error("Could not load project '" + directory.getName() + "'.");
						}
					}
				}
			}
		}

		return resultList;
	}

	/**
	 * This method uploads a given file as a new dependency. The file is stored
	 * within the lib-directory of this application. If a file with the same
	 * name does already exist, it will be replaces.
	 * 
	 * @param file
	 *            The file to be uploaded.
	 * @return The new dependency iff the uploading was sucesfull, null
	 *         otherwise.
	 */
	public MIDependency uploadDependency(final UploadedFile file) {
		final File depFile = new File(FileManager.LIB_DIR, file.getFileName());

		InputStream in = null;
		OutputStream out = null;
		try {
			/*
			 * Get the streams.
			 */
			in = file.getInputstream();
			out = new FileOutputStream(depFile);
			final byte[] buf = new byte[FileManager.BUF_SIZE];
			int count;

			/*
			 * Transfer the file.
			 */
			while ((count = in.read(buf)) != -1) {
				out.write(buf, 0, count);
			}

		} catch (final IOException ex) {
			FileManager.LOG.error("Could not transfer file '" + file.getFileName() + "'");
			return null;
		} finally {
			/*
			 * Try to make sure that the streams will be closed.
			 */
			try {
				if (in != null) {
					in.close();
				}
				if (out != null) {
					out.close();
				}
			} catch (final IOException ex) {
				// Ignore
			}
		}
		final MIDependency dependency = this.factory.createDependency();
		dependency.setFilePath(depFile.getAbsolutePath());
		return dependency;
	}

	/**
	 * Delivers the only instance of this class.
	 * 
	 * @return The singleton instance of this class.
	 */
	public static FileManager getInstance() {
		return FileManager.INSTANCE;
	}

	/**
	 * This method tries to load all available dependencies from the file system.
	 * 
	 * @return A list containing all available dependencies. If there are no files, an empty list will be returned.
	 */
	public List<MIDependency> loadAllDependencies() {
		final List<MIDependency> resultList = new ArrayList<MIDependency>();
		/*
		 * Try to get all files within the library directory.
		 */
		final File[] files = new File(FileManager.LIB_DIR).listFiles();
		if (files != null) {
			for (final File file : files) {
				if (file.isFile()) {
					if (file.getName().endsWith(FileManager.JAR_EXTENSION)) {
						final MIDependency dependency = this.factory.createDependency();
						dependency.setFilePath(file.getAbsolutePath());
						resultList.add(dependency);
					}
				}
			}
		}

		return resultList;
	}

	/**
	 * Tries to delete a given dependency.
	 * 
	 * @param dependency
	 *            The dependency to be removed.
	 * @return true iff the dependency exists and the removal was succesful.
	 */
	public boolean deleteDependency(final MIDependency dependency) {
		final File file = new File(dependency.getFilePath());
		if (file.isFile()) {
			return file.delete();
		}
		return false;
	}
}
